#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_naobot
----------------------------------

Tests for `naobot` module.
"""


class TestNaobot(object):

    def set_method(self):
        pass

    def test_something(self):
        pass

    def teardown_method(self, _):
        pass
